"""Labels."""

from __future__ import annotations


class Label(str):
    """Class representing labels for concepts."""

    HINT_SEP = ";"
    SPELLING_ALTERNATIVES_SEP = "|"

    def __eq__(self, other: object) -> bool:
        """Ignore hints when determining equality."""
        return self.without_hint == Label(other).without_hint

    def __ne__(self, other: object) -> bool:
        """Return whether the labels are not equal."""
        return not self == other

    @property
    def spelling_alternatives(self) -> Labels:
        """Extract the spelling alternatives from the label."""
        return label_factory(self.without_hint.split(self.SPELLING_ALTERNATIVES_SEP))

    @property
    def hint(self) -> str:
        """Return the label hint, if any."""
        return self.split(self.HINT_SEP)[1] if self.HINT_SEP in self else ""

    @property
    def without_hint(self) -> str:
        """Return the label without the hint."""
        return self.split(self.HINT_SEP)[0]


Labels = tuple[Label, ...]


def label_factory(string: str | list[str]) -> Labels:
    """Instantiate the labels from a string or list of strings."""
    labels = string if isinstance(string, list) else [string]
    return tuple(Label(label) for label in labels if not label.startswith("("))


def meaning_factory(string: str | list[str]) -> Labels:
    """Instantiate the meanings from a string or list of strings."""
    meanings = string if isinstance(string, list) else [string]
    return (Label(meanings[0].removeprefix("(").removesuffix(")")),) if meanings else Labels()



"""Meta data about this application."""

import pathlib
from importlib.metadata import metadata, version

import requests

from toisto.model.language import Language

_metadata = metadata("Toisto")
NAME = _metadata["name"]
SUMMARY = _metadata["summary"]
_homepage_url = _metadata["Project-URL"].split(", ")[1]
README_URL = f"{_homepage_url}/blob/main/README.md"
CHANGELOG_URL = [url for url in _metadata.get_all("Project-URL") if "Changelog" in url][0].split(", ")[1]
TAGS_API_URL = "https://api.github.com/repos/fniessink/toisto/tags"

VERSION = version(NAME)
BUILT_IN_LANGUAGES = [Language("en"), Language("fi"), Language("nl")]

# File locations
_data_folder = pathlib.Path(__file__).parent.parent
TOPIC_JSON_FILES = sorted((_data_folder / "topics").glob("*.json"))
LANGUAGES_FILE = _data_folder / "languages" / "iana-language-subtag-registry.txt"


def get_progress_filepath(target_language: Language) -> pathlib.Path:
    """Return the filename of the progress file for the specified target language."""
    return pathlib.Path.home() / f".{NAME.lower()}-progress-{target_language}.json"


TOPICS = [json_file.stem for json_file in TOPIC_JSON_FILES]


def latest_version() -> str | None:
    """Return the latest version."""
    try:
        return requests.get(TAGS_API_URL, timeout=2).json()[0]["name"]
    except requests.RequestException:
        return None



"""All languages as defined in the IANA language subtag registry."""

from toisto.persistence.iana_language_subtag_registry import load_languages

IANA_LANGUAGE_SUBTAG_REGISTRY_URL = "https://www.iana.org/assignments/language-subtag-registry"
ALL_LANGUAGES = load_languages()



"""Config file parser."""

import sys
from argparse import ArgumentParser
from collections.abc import Iterable
from configparser import ConfigParser, Error
from pathlib import Path
from typing import NoReturn, get_args

from toisto.model.language.cefr import CommonReferenceLevel
from toisto.model.language.iana_language_subtag_registry import ALL_LANGUAGES

# The schema for the config file. Top-level keys are sections, values are a dict per option with the key being the
# option name and the value being a tuple of a specifier and the allowed option values.
CONFIG_SCHEMA: dict[str, dict[str, tuple[str, Iterable]]] = dict(
    languages=dict(
        target=("one of", ALL_LANGUAGES.keys()),
        source=("one of", ALL_LANGUAGES.keys()),
        levels=("one or more of", get_args(CommonReferenceLevel)),
    ),
    commands=dict(mp3player=("any", [])),
)


class ConfigSchemaValidator:
    """Class to validate a config against the schema."""

    def __init__(self, config_parser: ConfigParser, argument_parser: ArgumentParser, config_filename: Path) -> None:
        self._argument_parser = argument_parser
        self._config_parser = config_parser
        self._config_filename = config_filename

    def _error(self, message: str) -> NoReturn:
        """Report the error and exit."""
        self._argument_parser.error(f"While reading from '{self._config_filename}': {message}")

    def validate(self) -> None | NoReturn:  # type: ignore[return]
        """Validate the config file against the schema."""
        for section in self._config_parser.sections():
            self._validate_section(section)

    def _validate_section(self, section: str) -> None | NoReturn:  # type: ignore[return]
        """Validate the section, including its options."""
        if section not in (allowed_sections := CONFIG_SCHEMA.keys()):
            self._error(f"unknown section '{section}'. Allowed sections are: {', '.join(allowed_sections)}.")
        for option in self._config_parser[section]:
            self._validate_option(section, option)

    def _validate_option(self, section: str, option: str) -> None | NoReturn:  # type: ignore[return]
        """Validate the option, including its value(s)."""
        if option not in (allowed_options := CONFIG_SCHEMA[section].keys()):
            self._error(
                f"unknown option '{option}' in section '{section}'. Allowed options are: {', '.join(allowed_options)}.",
            )
        specifier, allowed_values = CONFIG_SCHEMA[section][option]
        if specifier == "any":
            return  # type: ignore[return-value]
        value = self._config_parser.get(section, option)
        values = [value] if specifier == "one of" else value.split()
        if any(value for value in values if value not in allowed_values):
            self._error(
                f"unknown value '{value}' for option '{option}' in section '{section}'. "
                f"Allowed values are {specifier}: {', '.join(allowed_values)}.",
            )


def read_config(argument_parser: ArgumentParser) -> ConfigParser | NoReturn:
    """Read the config file, validate it, and exit with an error message if it doesn't pass."""
    parser = ConfigParser()
    config_filename = Path("~/.toisto.cfg").expanduser()
    try:
        with config_filename.open("r", encoding="utf-8") as config_file:
            parser.read_file(config_file)
    except FileNotFoundError:
        pass
    except (OSError, Error) as reason:
        argument_parser.error(str(reason))
    ConfigSchemaValidator(parser, argument_parser, config_filename).validate()
    _add_defaults(parser)
    return parser


def default_config() -> ConfigParser:
    """Return the default configuration."""
    parser = ConfigParser()
    _add_defaults(parser)
    return parser


def _add_defaults(parser: ConfigParser) -> None:
    """Add the default configuration to the parser."""
    if parser.get("commands", "mp3player", fallback=None) is None:
        if "commands" not in parser.sections():
            parser.add_section("commands")
        defaults = dict(darwin="afplay", linux="mpg123 --quiet")
        parser["commands"]["mp3player"] = defaults.get(sys.platform, "playsound")



"""Easter egg."""

import time
import webbrowser
from argparse import ArgumentParser

from toisto.ui.text import console


def show_easter_egg(argument_parser: ArgumentParser) -> None:
    """Show the easter egg."""
    console.print("🎉🐇🐣 Happy Easter! 🎉🐇🐣\n")
    time.sleep(1.0)
    console.print("Here are some ideas to do")
    time.sleep(1.0)
    console.print("over the Easter weekend")
    time.sleep(1.0)
    console.print("instead of practicing a language...")
    time.sleep(2.0)
    webbrowser.open("https://nl.pinterest.com/melliejacobs/pasen-paasei/")
    argument_parser.exit()



"""Load concepts from topic files and generate quizzes."""

import pathlib
from argparse import ArgumentParser
from typing import NoReturn

from ..metadata import NAME, TOPIC_JSON_FILES
from ..model.language import Language
from ..model.language.cefr import CommonReferenceLevel
from ..model.language.concept import ConceptId, ConceptIds
from ..model.language.concept_factory import ConceptFactory
from ..model.quiz.quiz_factory import QuizFactory
from ..model.quiz.topic import Topic, Topics
from .json_file import load_json


def load_topics(  # noqa: PLR0913
    target_language: Language,
    source_language: Language,
    levels: list[CommonReferenceLevel],
    builtin_topics_to_load: list[str],
    topic_files_to_load: list[str],
    argument_parser: ArgumentParser,
) -> Topics | NoReturn:
    """Load the topics with the concepts and generate the quizzes."""
    topics = set()
    topic_files: list[pathlib.Path] = []
    if builtin_topics_to_load or topic_files_to_load:
        topic_files.extend(topic for topic in TOPIC_JSON_FILES if topic.stem in builtin_topics_to_load)
        topic_files.extend(pathlib.Path(topic) for topic in topic_files_to_load)
    else:
        topic_files.extend(TOPIC_JSON_FILES)
    quiz_factory = QuizFactory(target_language, source_language)
    registry = ConceptIdRegistry(argument_parser)
    for topic_file in topic_files:
        concepts = []
        topic_quizzes = set()
        try:
            for concept_key, concept_value in load_json(topic_file).items():
                concept = ConceptFactory(concept_key, concept_value).create_concept()
                if levels and concept.level not in levels:
                    continue
                concepts.append(concept)
                topic_quizzes.update(quiz_factory.create_quizzes(concept))
        except Exception as reason:  # noqa: BLE001
            argument_parser.error(f"{NAME} cannot read topic {topic_file}: {reason}.\n")
        concept_ids = tuple({quiz.concept.base_concept.concept_id for quiz in topic_quizzes})
        registry.check_concept_ids(concept_ids, topic_file)
        registry.register_concept_ids(concept_ids, topic_file)
        topics.add(Topic(topic_file.stem, tuple(concepts), topic_quizzes))
    return Topics(topics)


class ConceptIdRegistry:
    """Registry to check the uniqueness of concept identifiers across topic files."""

    def __init__(self, argument_parser: ArgumentParser) -> None:
        self.argument_parser = argument_parser
        self.topic_files_by_concept_id: dict[ConceptId, pathlib.Path] = {}

    def check_concept_ids(self, concept_ids: ConceptIds, topic_file: pathlib.Path) -> None:
        """Check that the concept ids are unique."""
        for concept_id in concept_ids:
            self._check_concept_id(concept_id, topic_file)

    def _check_concept_id(self, concept_id: ConceptId, topic_file: pathlib.Path) -> None:
        """Check that the concept id is unique."""
        if concept_id in self.topic_files_by_concept_id:
            other_topic_file = self.topic_files_by_concept_id[concept_id]
            self.argument_parser.error(
                f"{NAME} cannot read topic {topic_file}: concept identifier '{concept_id}' also occurs in topic "
                f"'{other_topic_file}'.\nConcept identifiers must be unique across topic files.\n",
            )

    def register_concept_ids(self, concept_ids: ConceptIds, topic_file: pathlib.Path) -> None:
        """Register the concept ids."""
        for concept_id in concept_ids:
            self.topic_files_by_concept_id[concept_id] = topic_file



"""Quiz classes."""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import Literal, cast, get_args

from ..language import Language
from ..language.concept import Concept
from ..language.grammar import GrammaticalCategory
from ..language.iana_language_subtag_registry import ALL_LANGUAGES
from ..language.label import Label, Labels
from .match import match

TranslationQuizType = Literal["read", "write"]
ListenQuizType = Literal["listen"]
SemanticQuizType = Literal["antonym"]
GrammaticalQuizType = Literal[
    "pluralize",
    "singularize",
    "diminutize",
    "masculinize",
    "feminize",
    "neuterize",
    "give positive degree",
    "give comparative degree",
    "give superlative degree",
    "give first person",
    "give second person",
    "give third person",
    "give infinitive",
    "give present tense",
    "give past tense",
    "make declarative",
    "make interrogative",
    "affirm",
    "negate",
]
QuizType = Literal[TranslationQuizType, ListenQuizType, SemanticQuizType, GrammaticalQuizType]
GRAMMATICAL_QUIZ_TYPES: dict[GrammaticalCategory, GrammaticalQuizType] = {
    "plural": "pluralize",
    "singular": "singularize",
    "diminutive": "diminutize",
    "male": "masculinize",
    "female": "feminize",
    "neuter": "neuterize",
    "positive degree": "give positive degree",
    "comparative degree": "give comparative degree",
    "superlative degree": "give superlative degree",
    "first person": "give first person",
    "second person": "give second person",
    "third person": "give third person",
    "infinitive": "give infinitive",
    "present tense": "give present tense",
    "past tense": "give past tense",
    "declarative": "make declarative",
    "interrogative": "make interrogative",
    "affirmative": "affirm",
    "negative": "negate",
}
QUIZ_TYPE_GRAMMATICAL_CATEGORIES = {value: key for key, value in GRAMMATICAL_QUIZ_TYPES.items()}
INSTRUCTIONS: dict[Literal[TranslationQuizType, ListenQuizType, SemanticQuizType], str] = dict(
    read="Translate into",
    write="Translate into",
    listen="Listen and write in",
    antonym="Give the [underline]antonym[/underline] in",
)


@dataclass(frozen=True)
class Quiz:
    """Class representing a quiz."""

    concept: Concept
    question_language: Language
    answer_language: Language
    _question: Label
    _answers: Labels
    quiz_types: tuple[QuizType, ...] = ("read",)
    blocked_by: tuple[Quiz, ...] = ()
    _meanings: Labels = Labels()

    def __hash__(self) -> int:
        """Return a hash using the same attributes as used for testing equality."""
        return hash(self.key)

    def __eq__(self, other: object) -> bool:
        """Return whether this quiz is equal to the other."""
        return self.key == other.key if isinstance(other, self.__class__) else False

    def __ne__(self, other: object) -> bool:
        """Return whether this quiz is not equal to the other."""
        return not self == other

    @cached_property
    def key(self) -> str:
        """Return a string version of the quiz that can be used as key in the progress dict."""
        concept_id = self.concept.base_concept.concept_id
        quiz_types = "+".join(self.quiz_types)
        return f"{concept_id}:{self.question_language}:{self.answer_language}:{self.question}:{quiz_types}"

    def is_correct(self, guess: Label) -> bool:
        """Return whether the guess is correct."""
        return match(guess, *self.answers)

    @cached_property
    def question(self) -> Label:
        """Return the first spelling alternative of the question."""
        return self._question.spelling_alternatives[0]

    @property
    def answer(self) -> Label:
        """Return the first spelling alternative of the first answer."""
        return self._answers[0].spelling_alternatives[0]

    @property
    def answers(self) -> Labels:
        """Return all answers."""
        answers = [answer.spelling_alternatives for answer in self._answers]
        return cast(Labels, tuple(chain(*answers)))

    @property
    def meanings(self) -> Labels:
        """Return the first spelling alternative of the meanings."""
        return Labels(meaning.spelling_alternatives[0] for meaning in self._meanings)

    def other_answers(self, guess: Label) -> Labels:
        """Return the answers not equal to the guess."""
        if self.quiz_types[0] in get_args(ListenQuizType):
            return Labels()  # Other answers doesn't make sense if the user has to type what is spoken
        return tuple(answer for answer in self.answers if not match(guess, answer))

    def instruction(self) -> str:
        """Generate the quiz instruction."""
        hint = self._question.hint
        hint_text = f" ({hint})" if self.question_language != self.answer_language and hint else ""
        if self.quiz_types[0] in get_args(GrammaticalQuizType):
            categories = " ".join(
                QUIZ_TYPE_GRAMMATICAL_CATEGORIES[cast(GrammaticalQuizType, quiz_type)] for quiz_type in self.quiz_types
            )
            instruction_text = f"Give the [underline]{categories}[/underline] in"
        else:
            quiz_type = cast(Literal[TranslationQuizType, ListenQuizType, SemanticQuizType], self.quiz_types[0])
            instruction_text = INSTRUCTIONS[quiz_type]
        return f"{instruction_text} {ALL_LANGUAGES[self.answer_language]}{hint_text}"

    def is_blocked_by(self, quizzes: Quizzes) -> bool:
        """Return whether this quiz should come after any of the given quizzes."""
        return bool(set(self.blocked_by) & quizzes)


Quizzes = set[Quiz]
